package io.tswf.minecraft.blockventure.core.api.registry.lookup

import com.google.auto.service.AutoService
import com.mojang.serialization.Lifecycle
import groovy.transform.CompileStatic
import net.minecraft.registry.*
import net.minecraft.structure.processor.StructureProcessorLists
import net.minecraft.world.gen.carver.ConfiguredCarvers
import net.minecraft.world.gen.feature.ConfiguredFeatures
import net.minecraft.world.gen.feature.PlacedFeatures

import java.util.concurrent.ConcurrentHashMap

@CompileStatic
class StaticDynamicRegistryLookup {

    static Map<RegistryKey, RegistryBuilder.BootstrapFunction> bootstrapRegistries = new ConcurrentHashMap<>()

    @Lazy
    static List<ILookupSetupListener> setupListeners = ServiceLoader.load(ILookupSetupListener)
            .toList()

    static DynamicRegistryManager getStaticRegistryManager() {
        return DynamicRegistryManager.of(Registries.REGISTRIES)
    }

    /**
     * Собирает статические Registry и подпихивает по-возможности весь контент из датапаков, до какого можно достучаться.
     * @return
     */
    static RegistryWrapper.WrapperLookup createEnrichedRegistryLookup() {
        // Собираем все статические (доступные во время старта игры) и динамические (датапаки) registry

        // Начинаем со статических
        def allStaticRegistries = Registries.REGISTRIES.stream()
                .toList()

        def allStaticRegistryKeys = allStaticRegistries.map { it.key }

        // Затем выбираем динамические. Их будем оставлять пустыми по-умолчанию.
        // Для этого смотрим какие RegistryKey есть, вычленяем те, которых нет среди статических и создаем для них пустые SimpleRegistry
        def allDynamicRegistryKeys = RegistryKeys.fields
            .stream()
                // Чтобы не писать 100500 констант и не следить за их актуализацией - просто бегаем по всем полям внутри класса RegistryKeys, который хранит в себе все ванилла ключи и из полей-констант достаем ключи.
                //TODO: Возможно тут все-таки выбрать более красивый способ?
                .filter { it.modifierInfo.with { isPublic() && isStatic() && isFinal() } }
                .filter { it.type == RegistryKey }
                .map { (RegistryKey) it.get(null) }
                .filter { it !in allStaticRegistryKeys }

                // Каким-то неведомым образом тут иногда образуются дубликаты. Но потому как речь идет об уникальных ключах, то идентичные дубликаты можно отбросить в этих условиях.
                .distinct()
                .toList()


        // Формируем суммарный список всех Registry
        def allRegistries = new ArrayList<Registry>()
        allRegistries.addAll()

        // Добавляем в него заглушки для отсутствующих registry
        for (def dynamicRegistryKey : allDynamicRegistryKeys) {
            allRegistries << new SimpleRegistry<>(dynamicRegistryKey, Lifecycle.stable())
        }

        // Игнорируем те registry, которые будут загружены из датапаков. Иначе они будут дублироваться в DynamicRegistryManager и RegistryBuilder.
        allRegistries.removeIf { bootstrapRegistries.containsKey(it.key) }

        def dynRegistries = new DynamicRegistryManager.ImmutableImpl((List) allRegistries)
        def regBuilder = new RegistryBuilder()

        // Загружаем все registry, которые надо.
        for (def bootstrapEntry : bootstrapRegistries) {
            regBuilder.addRegistry(bootstrapEntry.key, bootstrapEntry.value)
        }

        return regBuilder.createWrapperLookup(dynRegistries)
    }

    static <T> void registerBootstrap(RegistryKey<? extends Registry<T>> registryRef, RegistryBuilder.BootstrapFunction<T> bootstrapFunction) {
        bootstrapRegistries[registryRef] = bootstrapFunction
    }

    static {
        for (def setupListener : setupListeners) {
            setupListener.handleLookupSetup()
        }
    }

    static interface ILookupSetupListener {
        void handleLookupSetup()
    }

    @AutoService(ILookupSetupListener)
    static class WorldGenLookupSetupListener implements ILookupSetupListener {

        @Override
        void handleLookupSetup() {
            registerBootstrap(RegistryKeys.PLACED_FEATURE, { PlacedFeatures.bootstrap(it) })
            registerBootstrap(RegistryKeys.CONFIGURED_CARVER, { ConfiguredCarvers.bootstrap(it) })
            registerBootstrap(RegistryKeys.CONFIGURED_FEATURE, { ConfiguredFeatures.bootstrap(it) })
            registerBootstrap(RegistryKeys.PROCESSOR_LIST, { StructureProcessorLists.bootstrap(it) })
        }
    }
}