package io.tswf.minecraft.blockventure.core.nbt.mapper

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import groovy.transform.Memoized
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor
import io.tswf.minecraft.blockventure.core.api.annotation.MapNBT
import io.tswf.minecraft.blockventure.core.api.nbt.INBTObjectMappingProvider
import io.tswf.minecraft.blockventure.core.api.nbt.IReflectiveNBTObject
import io.tswf.minecraft.blockventure.core.api.nbt.NBTObjectMapper
import io.tswf.minecraft.blockventure.core.exception.DuplicateNBTNamesException
import io.tswf.minecraft.blockventure.core.exception.InvalidMapToNbtAnnotationLocationException
import net.minecraft.nbt.NbtCompound
import net.minecraft.nbt.NbtElement

import java.lang.reflect.Type
import java.util.function.BiConsumer
import java.util.function.Function
import java.util.function.Supplier

@AutoService(INBTObjectMappingProvider)
@CompileStatic
class NBTReflectiveMappingProvider implements INBTObjectMappingProvider {

    @Override
    NbtElement mapObjectToNBT(Object obj, Type objectType) {
        def nbtFields = readNBTObjectFields(obj.class)
        def resultCompound = new NbtCompound()
        
        for (def nbtField : nbtFields) {
            def objectValue = nbtField.getter.apply(obj)
            if (objectType != null) {
                def nbtValue = NBTObjectMapper.mapToNbt(objectValue, nbtField.genericType)
                
                if (nbtValue == null) {
                    throw new IllegalStateException("Object '${objectValue}' not mapped!")
                }

                resultCompound.put(nbtField.nbtName, nbtValue)
            }
        }
        
        return resultCompound
    }

    @Override
    Object mapNBTToObject(NbtElement nbtElement, Type objectType, Object initialInstance) {
        def compoundNbt = (NbtCompound) nbtElement
        def objectInstance
        def objectClass = objectType.extractClass()

        if (initialInstance != null) {
            if (objectClass.isAssignableFrom(initialInstance.class)) {
                objectInstance = initialInstance
            } else {
                throw new IllegalArgumentException("Initial object '${initialInstance}' with class '${initialInstance.class.name}' is not extends '${objectClass.name}'!")
            }
        } else {
            objectInstance = objectClass.noArgsConstructor
                .ensureNotNull { new NoSuchMethodException("There is no default constructor to create instance of ${objectClass}!") }
                .newInstance()
        }

        def nbtFields = readNBTObjectFields(objectInstance.class)
        for (def nbtField : nbtFields) {
            def nbtElementForField = compoundNbt.get(nbtField.nbtName)
            if (nbtElementForField == null) {
                if (nbtField.rewriteIfNull) {
                    nbtField.setter.accept(objectInstance, nbtField.genericType.extractClass().defaultValue)
                }
            } else {
                def fieldInitialValue = nbtField.getter.apply(objectInstance)
                def nbtReadedValue = NBTObjectMapper.mapToObject(nbtElementForField, nbtField.genericType, fieldInitialValue)

                if (nbtReadedValue == null) {
                    throw new IllegalStateException("NBT field named '${nbtField.nbtName}' from ${objectInstance} not mapped!")
                }

                nbtField.setter.accept(objectInstance, nbtReadedValue)
            }
        }

        return objectInstance
    }

    @Override
    boolean canAcceptObjectType(Type objectType) {
        return IReflectiveNBTObject.isAssignableFrom(objectType.extractClass())
    }

    @Override
    boolean canAcceptNbtType(Class nbtType) {
        return nbtType == NbtCompound
    }
    
    @Memoized
    static List<NBTObjectField> readNBTObjectFields(Class classOfObject) {
        def result = classOfObject.allDeclaredFields.stream()
            .map {field ->
                field.getAnnotation(MapNBT)?.with { mapNbtAnnotation ->
                    def name = mapNbtAnnotation.name()
                    if (name.isNullOrBlank()) {
                        name = field.name
                    }
                    
                    if (field.modifierInfo.with {isStatic() || isFinal() || isTransient() }) {
                        throw new InvalidMapToNbtAnnotationLocationException("Field can not be static, final or transient", field)
                    }
                    
                    def rewriteIfNull = mapNbtAnnotation.rewriteIfNull()
                    
                    def getter = field.getters
                            .firstOrNull()
                            .ensureNotNull { new NoSuchMethodException("There is no getter for anotated with @${MapNBT.simpleName} field ${field}!") }

                    def setter = field.setters
                            .firstOrNull()
                            .ensureNotNull { new NoSuchMethodException("There is no setter for anotated with @${MapNBT.simpleName} field ${field}!") }


                    new NBTObjectField(
                            nbtName: name,
                            getter: { getter.invoke(it) },
                            setter: { self, val -> setter.invoke(self, val) },
                            genericType: field.genericType,
                            rewriteIfNull: rewriteIfNull
                    )
                }
            }
            .filter { it != null }
            .toList()
        
        def duplicateNames = result
            .groupBy { it.nbtName }
            .entrySet().stream()
                .filter { it.value.size() > 1 }
                .map {it.key }
                .toSet()
        
        if (duplicateNames.isNotEmpty()) {
            throw new DuplicateNBTNamesException(duplicateNames, classOfObject)
        }
        
        return result
    }
    
    @CheckedMapConstructor
    static class NBTObjectField {
        String nbtName
        // this и значение
        Function<Object, Object> getter
        BiConsumer<Object, Object> setter
        Type genericType
        
        boolean rewriteIfNull
    }
}
