package io.tswf.minecraft.blockventure.core.recipe

import groovy.transform.CompileStatic
import net.minecraft.recipe.Ingredient
import net.minecraft.util.collection.DefaultedList

@CompileStatic
class RecipeParser {

    static DefaultedList<Ingredient> parseShapelessRecipe(Object... recipeIngredientObjects) {
        return recipeIngredientObjects.stream()
            .map { RecipeIngredientObjectParser.parseIngredientObject(it) }
            .toList()
            .with {
                def result = DefaultedList.ofSize(it.size(), Ingredient.EMPTY)

                it.eachWithIndex { ingredient, idx ->
                    result[idx] = ingredient
                }

                return result
            }
    }

    static Tuple3<DefaultedList<Ingredient>, Integer, Integer> parseShapedRecipePattern(Object... recipeLinesAndIngredients) {
        def recipePatternLines = recipeLinesAndIngredients.stream()
                .takeWhile { (it instanceof String || it instanceof GString) && it.length() > 1 }
                .map { it.toString() }
                .toList()

        if (recipePatternLines.isNullOrEmpty()) {
            throw new IllegalArgumentException("There is no recipe lines in ${Arrays.toString(recipeLinesAndIngredients)} passed!")
        }

        def longestRecipePatternLine = recipePatternLines.stream()
                .max(Comparator.comparingInt { it.toString().length() })
                // Выше проверили, что список строк не пустой. Так что тут всегда будет значение
                .get()

        for (def recipeLine : recipePatternLines) {
            if (recipeLine.length() != longestRecipePatternLine.length()) {
                throw new IllegalArgumentException("Invalid recipe pattern line size! Expected ${longestRecipePatternLine.length()}, but was ${recipeLine.length()} for line '${recipeLine}'. All recipe pattern lines must have equal length (as '${longestRecipePatternLine}')!")
            }
        }

        def recipeW = longestRecipePatternLine.length()
        def recipeH = recipePatternLines.size()

        def recipePattern = recipePatternLines.join("")

        def ingredients = recipeLinesAndIngredients.stream()
            .skip(recipePatternLines.size())
            .toArray()

        if (ingredients.isNullOrEmpty()) {
            throw new IllegalArgumentException("No ingredient actors passed!")
        }

        if (ingredients.length % 2 != 0) {
            throw new IllegalArgumentException("There is no ingredient for actor '${ingredients.last()}'!")
        }

        def actorToIngredient = new HashMap<String, Ingredient>()
        for (def actorAndIngredientPair : ingredients.collate(2)) {
            def actor = actorAndIngredientPair[0]
            def ingredientObject = actorAndIngredientPair[1]

            actor.ensureNotNull { new IllegalArgumentException("There is null actor for ingredient object '${ingredientObject}'") }
            ingredientObject.ensureNotNull { new IllegalArgumentException("There is null ingredient object for actor '${actor}'") }

            if (actor !instanceof String) {
                throw new IllegalArgumentException("Actor '${actor}' must be a string, not a ${actor.class.name}!")
            }

            def actorString = (String) actor

            if (actorString.length() != 1) {
                throw new IllegalArgumentException("Actor's '${actorString}' length must be a 1, not a ${actorString.length()}!")
            }

            def parsedIngredient = RecipeIngredientObjectParser.parseIngredientObject(ingredientObject)
                    .ensureNotNull { new IllegalArgumentException("There is invalid ingredient object for actor '${actorString}': ${ingredientObject}") }

            actorToIngredient[actorString] = parsedIngredient
        }

        def recipeIngredients = DefaultedList.ofSize(recipePattern.length(), Ingredient.EMPTY)
        for (def patternActorIndex = 0; patternActorIndex < recipePattern.length(); patternActorIndex ++) {
            def patternActor = recipePattern.charAt(patternActorIndex).toString()

            if (!patternActor.isNullOrBlank()) {
                def patternIngredient = actorToIngredient[patternActor]
                        .ensureNotNull { new IllegalArgumentException("There is unknown recipe actor: '${patternActor}'") }

                recipeIngredients.set(patternActorIndex, patternIngredient)
            }
        }

        return new Tuple3<DefaultedList<Ingredient>, Integer, Integer>(recipeIngredients, recipeW, recipeH)
    }
}
