package io.tswf.minecraft.blockventure.core.recipe.parser

import com.google.auto.service.AutoService
import com.google.gson.JsonObject
import groovy.transform.CompileStatic
import io.tswf.minecraft.blockventure.core.recipe.IRecipeIngredientObjectParseProvider
import io.tswf.minecraft.blockventure.core.recipe.RecipeIngredientObjectParser
import io.tswf.minecraft.blockventure.core.tag.TagParser
import net.minecraft.item.BlockItem
import net.minecraft.item.ItemStack
import net.minecraft.recipe.Ingredient
import net.minecraft.registry.Registries
import net.minecraft.registry.RegistryKey
import net.minecraft.registry.tag.TagKey

@AutoService(IRecipeIngredientObjectParseProvider)
@CompileStatic
class TagIngredientObjectParseProvider implements IRecipeIngredientObjectParseProvider{

    public static final String TAG_PREFIX = "tag:"

    Ingredient parseTagIngrediet(TagKey tag) {
        return new TagIngredient(tag)
    }

    @Override
    Ingredient parseIngredientObject(Object ingredientObject) {
        if (ingredientObject instanceof TagKey) {
            return parseTagIngrediet(ingredientObject)
        }

        if (ingredientObject instanceof String) {
            def strippedIngredientObject = ingredientObject.stripAll()
            if (strippedIngredientObject.startsWith(TAG_PREFIX)) {
                def tagString = strippedIngredientObject.substring(TAG_PREFIX.length()).stripAll()
                def tag = TagParser.parseTag(tagString)

                return parseTagIngrediet(tag)
            }
        }

        return null
    }
}

@CompileStatic
class TagIngredient extends Ingredient {

    public static final Set<RegistryKey> ALLOWED_TAGS_KEYS = [ Registries.BLOCK.key, Registries.ITEM.key ] as Set<RegistryKey>

    TagKey tag

    TagIngredient(TagKey tag) {
        super(Arrays.stream(new TagEntry(tag)))
        this.tag = tag
    }

    @CompileStatic
    static class TagEntry implements Ingredient.Entry {
        TagKey tag

        TagEntry(TagKey tag) {
            this.tag = tag
            validateTag(tag)
        }

        void validateTag(TagKey key) {
            def tagRegistry = key.registry()
            if (tagRegistry !in ALLOWED_TAGS_KEYS) {
                throw new IllegalArgumentException("There is tag registry '${tagRegistry}' from tag '${key}' is not supported by ${this.class.simpleName}! Only ${ALLOWED_TAGS_KEYS} supported!")
            }
        }

        @Override
        Collection<ItemStack> getStacks() {
            validateTag(tag)
            def tagRegistry = tag.registry()
            if (tagRegistry == Registries.BLOCK.key) {
                def blocks = Registries.BLOCK.iterateEntries(tag).toList()
                def blockItemsStacks = blocks.stream()
                    .map { BlockItem.BLOCK_ITEMS[it.get()] }
                    .filter { it != null }
                    .map { new ItemStack(it) }
                    .toList()

                return blockItemsStacks
            } else if (tagRegistry == Registries.ITEM.key) {
                return Registries.ITEM.iterateEntries(tag)
                        .map { new ItemStack(it.get()) }
            } else {
                throw new IllegalArgumentException("Invalid registry key for tag '${tag}': '${tagRegistry}'. Only blocks and items allowed!")
            }
        }


        @Override
        JsonObject toJson() {
            return new JsonObject()
        }
    }
}
